/** animation.dart --- Easily create animations and effects.
 *
 * Copyright (C) 2014  Miguel Guedes
 * 
 * Author: Miguel Guedes <miguel.a.guedes@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see http://www.gnu.org/licenses/.
 *
 * Comments:
 * 
 */


import 'dart:html';
import 'dart:async';
import 'package:observable/observable.dart';

import 'animator.dart';
import 'easing.dart';
import 'src/convert.dart';
import 'animation_exception.dart';


class Animation extends Observable
{
  static const HERTZ = 40;
  static const DEFAULT_DURATION = 400;
  static const DEFAULT_EASING = Easing.easeInQuad;
  
  
  Element   _element;
  String    _property;
  num       _targetValue;
  String    _units;
  function  _easing;
  int       _duration;          /* in milliseconds */
  num       _initialValue;
  Timer     _timer;
  Stopwatch _stopwatch;
  
  
  Animation(final Element   element,
            final String    property,
          [ final Duration  duration,
            final function  easing ] )
    : _element = element,
      _duration = duration == null ? DEFAULT_DURATION
                                   : duration.inMilliseconds,
      _easing = easing == null ? DEFAULT_EASING : easing
  {
    RegEx rx = new RegExp(r"^\s*([a-zA-Z\-]+):\s*(\d*(\.\d*)?)([a-zA-Z]+)?");
    Match m = rx.firstMatch(property);

    if(m == null)
      throw new AnimationException("Unable to extract name and target value from property");

    _property = m.group(1);
    _initialValue = Convert.parseNum(_element.getComputedStyle()
        .getPropertyValue(_property));

    _targetValue = Convert.parseNum(m.group(2)) - _initialValue;

    _units = m.group(4);
    if(_units == null)
      _units = "";
  }
  
  
  Animation start()
  {
    if(_timer != null)
      throw new AnimationException("Animation already started");
    
    _stopwatch = new Stopwatch()
      ..start();
    
    _timer = new Timer.periodic(
        new Duration(milliseconds: 1 / (HERTZ * 1000)),
        _tick);

    Animator.add(this);
    observer.dispatch("start");

    return this;
  }

  void stop()
  {
    if(_timer == null)
      throw new AnimationException("Animation not yet started");

    _stopwatch.stop();
    _timer.cancel();
    _timer = null;

    Animator.remove(this);
    observer.dispatch("end");
  }
  
  void _tick(final Timer timer)
  {
    int currentTime = _stopwatch.elapsedMilliseconds;

    if(currentTime > _duration)
      currentTime = _duration;

    final num currentValue = _easing(currentTime,
                                     _initialValue,
                                     _targetValue,
                                     _duration);

    _element.style.setProperty(_property,
        currentValue.toString() + _units);

    if(currentTime >= _duration)
      stop();
  }

  Element getElement() => _element;

  String getProperty() => _property;

  bool running() => _timer != null;
}